(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.MagicString = factory());
}(this, function () { 'use strict';

	function Chunk(start, end, content) {
		this.start = start;
		this.end = end;
		this.original = content;

		this.content = content;
		this.storeName = false;
		this.edited = false;
	}

	Chunk.prototype = {
		clone: function clone() {
			var chunk = new Chunk(this.start, this.end, this.original);
			chunk.content = this.content;
			chunk.storeName = this.storeName;
			chunk.edited = this.edited;

			return chunk;
		},
		edit: function edit(content, storeName) {
			this.content = content;
			this.storeName = storeName;

			this.edited = true;

			return this;
		},
		split: function split(index) {
			if (index === this.start) return this;

			var sliceIndex = index - this.start;

			var originalBefore = this.original.slice(0, sliceIndex);
			var originalAfter = this.original.slice(sliceIndex);

			this.original = originalBefore;

			var newChunk = new Chunk(index, this.end, originalAfter);
			this.end = index;

			if (this.edited) {
				if (this.content.length) throw new Error('Cannot split a chunk that has already been edited ("' + this.original + '")');

				// zero-length edited chunks are a special case (overlapping replacements)
				newChunk.edit('', false);
				this.content = '';
			} else {
				this.content = originalBefore;
			}

			return newChunk;
		}
	};

	var _btoa = void 0;

	if (typeof window !== 'undefined' && typeof window.btoa === 'function') {
		_btoa = window.btoa;
	} else if (typeof Buffer === 'function') {
		/* global Buffer */
		_btoa = function _btoa(str) {
			return new Buffer(str).toString('base64');
		};
	} else {
		throw new Error('Unsupported environment: `window.btoa` or `Buffer` should be supported.');
	}

	var btoa = _btoa;

	function SourceMap(properties) {
		this.version = 3;

		this.file = properties.file;
		this.sources = properties.sources;
		this.sourcesContent = properties.sourcesContent;
		this.names = properties.names;
		this.mappings = properties.mappings;
	}

	SourceMap.prototype = {
		toString: function toString() {
			return JSON.stringify(this);
		},
		toUrl: function toUrl() {
			return 'data:application/json;charset=utf-8;base64,' + btoa(this.toString());
		}
	};

	function guessIndent(code) {
		var lines = code.split('\n');

		var tabbed = lines.filter(function (line) {
			return (/^\t+/.test(line)
			);
		});
		var spaced = lines.filter(function (line) {
			return (/^ {2,}/.test(line)
			);
		});

		if (tabbed.length === 0 && spaced.length === 0) {
			return null;
		}

		// More lines tabbed than spaced? Assume tabs, and
		// default to tabs in the case of a tie (or nothing
		// to go on)
		if (tabbed.length >= spaced.length) {
			return '\t';
		}

		// Otherwise, we need to guess the multiple
		var min = spaced.reduce(function (previous, current) {
			var numSpaces = /^ +/.exec(current)[0].length;
			return Math.min(numSpaces, previous);
		}, Infinity);

		return new Array(min + 1).join(' ');
	}

	var charToInteger = {};
	var integerToChar = {};

	'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/='.split( '' ).forEach( function ( char, i ) {
		charToInteger[ char ] = i;
		integerToChar[ i ] = char;
	});

	function encode ( value ) {
		var result, i;

		if ( typeof value === 'number' ) {
			result = encodeInteger( value );
		} else {
			result = '';
			for ( i = 0; i < value.length; i += 1 ) {
				result += encodeInteger( value[i] );
			}
		}

		return result;
	}

	function encodeInteger ( num ) {
		var result = '', clamped;

		if ( num < 0 ) {
			num = ( -num << 1 ) | 1;
		} else {
			num <<= 1;
		}

		do {
			clamped = num & 31;
			num >>= 5;

			if ( num > 0 ) {
				clamped |= 32;
			}

			result += integerToChar[ clamped ];
		} while ( num > 0 );

		return result;
	}

	function encodeMappings(original, intro, chunks, hires, sourcemapLocations, sourceIndex, offsets, names) {
		var rawLines = [];

		var generatedCodeLine = intro.split('\n').length - 1;
		var rawSegments = rawLines[generatedCodeLine] = [];

		var originalCharIndex = 0;

		var generatedCodeColumn = 0;
		var sourceCodeLine = 0;
		var sourceCodeColumn = 0;

		function addSegmentsUntil(end) {
			var first = true;

			while (originalCharIndex < end) {
				if (hires || first || sourcemapLocations[originalCharIndex]) {
					rawSegments.push({
						generatedCodeLine: generatedCodeLine,
						generatedCodeColumn: generatedCodeColumn,
						sourceCodeLine: sourceCodeLine,
						sourceCodeColumn: sourceCodeColumn,
						sourceCodeName: -1,
						sourceIndex: sourceIndex
					});
				}

				if (original[originalCharIndex] === '\n') {
					sourceCodeLine += 1;
					sourceCodeColumn = 0;
					generatedCodeLine += 1;
					rawLines[generatedCodeLine] = rawSegments = [];
					generatedCodeColumn = 0;
				} else {
					sourceCodeColumn += 1;
					generatedCodeColumn += 1;
				}

				originalCharIndex += 1;
				first = false;
			}
		}

		for (var i = 0; i < chunks.length; i += 1) {
			var chunk = chunks[i];

			if (chunk.edited) {
				if (i > 0 || chunk.content.length) {
					rawSegments.push({
						generatedCodeLine: generatedCodeLine,
						generatedCodeColumn: generatedCodeColumn,
						sourceCodeLine: sourceCodeLine,
						sourceCodeColumn: sourceCodeColumn,
						sourceCodeName: chunk.storeName ? names.indexOf(chunk.original) : -1,
						sourceIndex: sourceIndex
					});
				}

				var lines = chunk.content.split('\n');
				var lastLine = lines.pop();

				if (lines.length) {
					generatedCodeLine += lines.length;
					rawLines[generatedCodeLine] = rawSegments = [];
					generatedCodeColumn = lastLine.length;
				} else {
					generatedCodeColumn += lastLine.length;
				}

				lines = chunk.original.split('\n');
				lastLine = lines.pop();

				if (lines.length) {
					sourceCodeLine += lines.length;
					sourceCodeColumn = lastLine.length;
				} else {
					sourceCodeColumn += lastLine.length;
				}
			} else {
				addSegmentsUntil(chunk.end);
			}

			originalCharIndex = chunk.end;
		}

		addSegmentsUntil(original.length);

		offsets.sourceIndex = offsets.sourceIndex || 0;
		offsets.sourceCodeLine = offsets.sourceCodeLine || 0;
		offsets.sourceCodeColumn = offsets.sourceCodeColumn || 0;
		offsets.sourceCodeName = offsets.sourceCodeName || 0;

		var encoded = rawLines.map(function (segments) {
			var generatedCodeColumn = 0;

			return segments.map(function (segment) {
				var arr = [segment.generatedCodeColumn - generatedCodeColumn, segment.sourceIndex - offsets.sourceIndex, segment.sourceCodeLine - offsets.sourceCodeLine, segment.sourceCodeColumn - offsets.sourceCodeColumn];

				generatedCodeColumn = segment.generatedCodeColumn;
				offsets.sourceIndex = segment.sourceIndex;
				offsets.sourceCodeLine = segment.sourceCodeLine;
				offsets.sourceCodeColumn = segment.sourceCodeColumn;

				if (~segment.sourceCodeName) {
					arr.push(segment.sourceCodeName - offsets.sourceCodeName);
					offsets.sourceCodeName = segment.sourceCodeName;
				}

				return encode(arr);
			}).join(',');
		}).join(';');

		return encoded;
	}

	function getRelativePath(from, to) {
		var fromParts = from.split(/[\/\\]/);
		var toParts = to.split(/[\/\\]/);

		fromParts.pop(); // get dirname

		while (fromParts[0] === toParts[0]) {
			fromParts.shift();
			toParts.shift();
		}

		if (fromParts.length) {
			var i = fromParts.length;
			while (i--) {
				fromParts[i] = '..';
			}
		}

		return fromParts.concat(toParts).join('/');
	}

	var toString = Object.prototype.toString;

	function isObject(thing) {
		return toString.call(thing) === '[object Object]';
	}

	var warned = false;

	function MagicString(string) {
		var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

		var chunk = new Chunk(0, string.length, string);

		Object.defineProperties(this, {
			original: { writable: true, value: string },
			outro: { writable: true, value: '' },
			intro: { writable: true, value: '' },
			chunks: { writable: true, value: [chunk] },
			moves: { writable: true, value: [] },
			filename: { writable: true, value: options.filename },
			indentExclusionRanges: { writable: true, value: options.indentExclusionRanges },
			sourcemapLocations: { writable: true, value: {} },
			storedNames: { writable: true, value: {} },
			indentStr: { writable: true, value: guessIndent(string) }
		});
	}

	MagicString.prototype = {
		addSourcemapLocation: function addSourcemapLocation(char) {
			this.sourcemapLocations[char] = true;
		},
		append: function append(content) {
			if (typeof content !== 'string') throw new TypeError('outro content must be a string');

			this.outro += content;
			return this;
		},
		clone: function clone() {
			var cloned = new MagicString(this.original, { filename: this.filename });

			cloned.chunks = this.chunks.map(function (chunk) {
				return chunk.clone();
			});

			if (this.indentExclusionRanges) {
				cloned.indentExclusionRanges = typeof this.indentExclusionRanges[0] === 'number' ? [this.indentExclusionRanges[0], this.indentExclusionRanges[1]] : this.indentExclusionRanges.map(function (range) {
					return [range.start, range.end];
				});
			}

			Object.keys(this.sourcemapLocations).forEach(function (loc) {
				cloned.sourcemapLocations[loc] = true;
			});

			return cloned;
		},
		generateMap: function generateMap(options) {
			options = options || {};

			var names = Object.keys(this.storedNames);

			return new SourceMap({
				file: options.file ? options.file.split(/[\/\\]/).pop() : null,
				sources: [options.source ? getRelativePath(options.file || '', options.source) : null],
				sourcesContent: options.includeContent ? [this.original] : [null],
				names: names,
				mappings: this.getMappings(options.hires, 0, {}, names)
			});
		},
		getIndentString: function getIndentString() {
			return this.indentStr === null ? '\t' : this.indentStr;
		},
		getMappings: function getMappings(hires, sourceIndex, offsets, names) {
			return encodeMappings(this.original, this.intro, this.chunks, hires, this.sourcemapLocations, sourceIndex, offsets, names);
		},
		indent: function indent(indentStr, options) {
			var pattern = /^[^\r\n]/gm;

			if (isObject(indentStr)) {
				options = indentStr;
				indentStr = undefined;
			}

			indentStr = indentStr !== undefined ? indentStr : this.indentStr || '\t';

			if (indentStr === '') return this; // noop

			options = options || {};

			// Process exclusion ranges
			var isExcluded = {};

			if (options.exclude) {
				var exclusions = typeof options.exclude[0] === 'number' ? [options.exclude] : options.exclude;
				exclusions.forEach(function (exclusion) {
					for (var i = exclusion[0]; i < exclusion[1]; i += 1) {
						isExcluded[i] = true;
					}
				});
			}

			var shouldIndentNextCharacter = options.indentStart !== false;
			var replacer = function replacer(match) {
				if (shouldIndentNextCharacter) return '' + indentStr + match;
				shouldIndentNextCharacter = true;
				return match;
			};

			this.intro = this.intro.replace(pattern, replacer);

			var chunkIndex = void 0;
			var charIndex = 0;

			for (chunkIndex = 0; chunkIndex < this.chunks.length; chunkIndex += 1) {
				// can't cache this.chunks.length, it may change
				var chunk = this.chunks[chunkIndex];
				var end = chunk.end;

				if (chunk.edited) {
					if (!isExcluded[charIndex]) {
						chunk.content = chunk.content.replace(pattern, replacer);

						if (chunk.content.length) {
							shouldIndentNextCharacter = chunk.content[chunk.content.length - 1] === '\n';
						}
					}
				} else {
					charIndex = chunk.start;

					while (charIndex < end) {
						if (!isExcluded[charIndex]) {
							var char = this.original[charIndex];

							if (char === '\n') {
								shouldIndentNextCharacter = true;
							} else if (char !== '\r' && shouldIndentNextCharacter) {
								shouldIndentNextCharacter = false;

								var indentation = new Chunk(charIndex, charIndex, '').edit(indentStr, false);
								var remainder = chunk.split(charIndex);

								if (charIndex === chunk.start) {
									this.chunks.splice(chunkIndex, 0, indentation);
									chunkIndex += 1;
								} else {
									this.chunks.splice(chunkIndex + 1, 0, indentation, remainder);
									chunkIndex += 2;
								}

								chunk = remainder;
							}
						}

						charIndex += 1;
					}
				}

				charIndex = chunk.end;
			}

			this.outro = this.outro.replace(pattern, replacer);

			return this;
		},
		insert: function insert(index, content) {
			if (typeof content !== 'string') throw new TypeError('inserted content must be a string');

			this._split(index);

			var next = this.chunks.findIndex(function (chunk) {
				return chunk.end > index;
			});
			if (! ~next) next = this.chunks.length;

			var newChunk = new Chunk(index, index, '').edit(content, false);

			this.chunks.splice(next, 0, newChunk);
			return this;
		},


		// get current location of character in original string
		locate: function locate(character) {
			throw new Error('magicString.locate is deprecated');
		},
		locateOrigin: function locateOrigin(character) {
			throw new Error('magicString.locateOrigin is deprecated');
		},
		move: function move(start, end, index) {
			if (index >= start && index <= end) throw new Error('Cannot move a selection inside itself');

			this._split(start);
			this._split(end);
			this._split(index);

			this.moves.push({ start: start, end: end, index: index });

			return this;
		},
		overwrite: function overwrite(start, end, content, storeName) {
			if (typeof content !== 'string') {
				throw new TypeError('replacement content must be a string');
			}

			this._split(start);
			this._split(end);

			if (storeName) {
				var original = this.original.slice(start, end);
				this.storedNames[original] = true;
			}

			var firstIndex = this.chunks.findIndex(function (chunk) {
				return chunk.start === start && chunk.original.length;
			});
			var lastIndex = this.chunks.findIndex(function (chunk) {
				return chunk.start === end;
			});
			if (! ~firstIndex) firstIndex = this.chunks.length;
			if (! ~lastIndex) lastIndex = this.chunks.length;

			var newChunk = new Chunk(start, end, this.original.slice(start, end)).edit(content, storeName);

			this.chunks.splice(firstIndex, lastIndex - firstIndex, newChunk);
			return this;
		},
		prepend: function prepend(content) {
			if (typeof content !== 'string') throw new TypeError('outro content must be a string');

			this.intro = content + this.intro;
			return this;
		},
		remove: function remove(start, end) {
			while (start < 0) {
				start += this.original.length;
			}while (end < 0) {
				end += this.original.length;
			}if (start === end) return this;

			if (start < 0 || end > this.original.length) throw new Error('Character is out of bounds');
			if (start > end) throw new Error('end must be greater than start');

			var firstIndex = this.chunks.findIndex(function (chunk) {
				return chunk.start <= start && chunk.end > start;
			});
			var chunk = this.chunks[firstIndex];

			// if the chunk contains `start`, split
			if (chunk.start < start) {
				if (chunk.edited && chunk.content.length) throw new Error('nope');
				this._split(start);
				firstIndex += 1;
			}

			var lastIndex = this.chunks.findIndex(function (chunk) {
				return chunk.start < end && chunk.end >= end;
			});
			chunk = this.chunks[lastIndex];

			// if the chunk contains `end`, split
			if (chunk.start < end) {
				if (chunk.edited && chunk.content.length) throw new Error('nope');
				this._split(end);
			}

			lastIndex += 1;

			var newChunk = new Chunk(start, end, this.original.slice(start, end)).edit('', false);
			this.chunks.splice(firstIndex, lastIndex - firstIndex, newChunk);

			return this;
		},
		replace: function replace(start, end, content) {
			if (!warned) {
				console.warn('magicString.replace(...) is deprecated. Use magicString.overwrite(...) instead');
				warned = true;
			}

			return this.overwrite(start, end, content);
		},
		slice: function slice(start) {
			var end = arguments.length <= 1 || arguments[1] === undefined ? this.original.length : arguments[1];

			while (start < 0) {
				start += this.original.length;
			}while (end < 0) {
				end += this.original.length;
			}var result = '';

			// TODO handle moves

			for (var i = 0; i < this.chunks.length; i += 1) {
				var chunk = this.chunks[i];

				if (chunk.end <= start) continue;
				if (chunk.start >= end) break;

				if (chunk.start < start || chunk.end > end) {
					if (chunk.edited) throw new Error('Cannot use replaced characters (' + start + ', ' + end + ') as slice anchors');

					var sliceStart = Math.max(start - chunk.start, 0);
					var sliceEnd = Math.min(chunk.content.length - (chunk.end - end), chunk.content.length);

					result += chunk.content.slice(sliceStart, sliceEnd);
				} else {
					result += chunk.content;
				}
			}

			return result;
		},
		snip: function snip(start, end) {
			var clone = this.clone();
			clone.remove(0, start);
			clone.remove(end, clone.original.length);

			return clone;
		},
		_sortChunks: function _sortChunks() {
			var chunks = this.chunks.slice();

			// TODO there must be a better way than this...

			this.moves.forEach(function (move, i) {
				var firstIndex = chunks.findIndex(function (chunk) {
					return chunk.start === move.start;
				});
				var lastIndex = chunks.findIndex(function (chunk, i) {
					return i >= firstIndex && chunk.end === move.end;
				}) + 1;
				if (!lastIndex) lastIndex = chunks.length;

				var insertionIndex = chunks.findIndex(function (chunk) {
					return chunk.start === move.index;
				});
				if (! ~insertionIndex) insertionIndex = chunks.length;
				var num = lastIndex - firstIndex;

				if (firstIndex < insertionIndex) insertionIndex -= num;

				var toMove = chunks.splice(firstIndex, num);
				chunks.splice.apply(chunks, [insertionIndex, 0].concat(toMove));
			});

			return chunks;
		},
		_split: function _split(index) {
			// TODO bisect
			for (var i = 0; i < this.chunks.length; i += 1) {
				var chunk = this.chunks[i];
				if (chunk.start === index || chunk.end === index) return;

				if (chunk.start < index && chunk.end > index) {
					var newChunk = chunk.split(index);
					this.chunks.splice(i + 1, 0, newChunk);
					return;
				}
			}
		},
		toString: function toString() {
			return this.intro + this._sortChunks().map(function (chunk) {
				return chunk.content;
			}).join('') + this.outro;
		},
		trimLines: function trimLines() {
			return this.trim('[\\r\\n]');
		},
		trim: function trim(charType) {
			return this.trimStart(charType).trimEnd(charType);
		},
		trimEnd: function trimEnd(charType) {
			var rx = new RegExp((charType || '\\s') + '+$');

			this.outro = this.outro.replace(rx, '');
			if (this.outro.length) return this;

			var charIndex = this.original.length;
			var i = this.chunks.length;

			while (i--) {
				var chunk = this.chunks[i];

				if (charIndex > chunk.end) {
					var _slice = this.original.slice(chunk.end, charIndex);

					var _match = rx.exec(_slice);
					if (_match) {
						this.chunk(charIndex - _match[0].length, charIndex, '');
					}

					if (!_match || _match[0].length < _slice.length) {
						// there is non-whitespace after the chunk
						return this;
					}
				}

				chunk.content = chunk.content.replace(rx, '');
				if (chunk.content) return this;

				charIndex = chunk.start;
			}

			var slice = this.original.slice(0, charIndex);

			var match = rx.exec(slice);
			if (match) this.chunk(charIndex - match[0].length, charIndex, '');

			return this;
		},
		trimStart: function trimStart(charType) {
			var rx = new RegExp('^' + (charType || '\\s') + '+');

			this.intro = this.intro.replace(rx, '');
			if (this.intro.length) return this;

			var charIndex = 0;

			for (var i = 0; i < this.chunks.length; i += 1) {
				var chunk = this.chunks[i];

				if (charIndex < chunk.start) {
					var _slice2 = this.original.slice(charIndex, chunk.start);

					var _match2 = rx.exec(_slice2);
					if (_match2) this.chunk(charIndex, charIndex + _match2[0].length, '');

					if (!_match2 || _match2[0].length < _slice2.length) {
						// there is non-whitespace before the chunk
						return this;
					}
				}

				chunk.content = chunk.content.replace(rx, '');
				if (chunk.content) return this;

				charIndex = chunk.end;
			}

			var slice = this.original.slice(charIndex, this.original.length);

			var match = rx.exec(slice);
			if (match) this.chunk(charIndex, charIndex + match[0].length, '');

			return this;
		}
	};

	var hasOwnProp = Object.prototype.hasOwnProperty;

	function Bundle() {
		var options = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

		this.intro = options.intro || '';
		this.separator = options.separator !== undefined ? options.separator : '\n';

		this.sources = [];

		this.uniqueSources = [];
		this.uniqueSourceIndexByFilename = {};
	}

	Bundle.prototype = {
		addSource: function addSource(source) {
			if (source instanceof MagicString) {
				return this.addSource({
					content: source,
					filename: source.filename,
					separator: this.separator
				});
			}

			if (!isObject(source) || !source.content) {
				throw new Error('bundle.addSource() takes an object with a `content` property, which should be an instance of MagicString, and an optional `filename`');
			}

			['filename', 'indentExclusionRanges', 'separator'].forEach(function (option) {
				if (!hasOwnProp.call(source, option)) source[option] = source.content[option];
			});

			if (source.separator === undefined) {
				// TODO there's a bunch of this sort of thing, needs cleaning up
				source.separator = this.separator;
			}

			if (source.filename) {
				if (!hasOwnProp.call(this.uniqueSourceIndexByFilename, source.filename)) {
					this.uniqueSourceIndexByFilename[source.filename] = this.uniqueSources.length;
					this.uniqueSources.push({ filename: source.filename, content: source.content.original });
				} else {
					var uniqueSource = this.uniqueSources[this.uniqueSourceIndexByFilename[source.filename]];
					if (source.content.original !== uniqueSource.content) {
						throw new Error('Illegal source: same filename (' + source.filename + '), different contents');
					}
				}
			}

			this.sources.push(source);
			return this;
		},
		append: function append(str, options) {
			this.addSource({
				content: new MagicString(str),
				separator: options && options.separator || ''
			});

			return this;
		},
		clone: function clone() {
			var bundle = new Bundle({
				intro: this.intro,
				separator: this.separator
			});

			this.sources.forEach(function (source) {
				bundle.addSource({
					filename: source.filename,
					content: source.content.clone(),
					separator: source.separator
				});
			});

			return bundle;
		},
		generateMap: function generateMap(options) {
			var _this = this;

			var offsets = {};

			var names = [];
			this.sources.forEach(function (source) {
				Object.keys(source.content.storedNames).forEach(function (name) {
					if (! ~names.indexOf(name)) names.push(name);
				});
			});

			var encoded = getSemis(this.intro) + this.sources.map(function (source, i) {
				var prefix = i > 0 ? getSemis(source.separator) || ',' : '';
				var mappings = void 0;

				// we don't bother encoding sources without a filename
				if (!source.filename) {
					mappings = getSemis(source.content.toString());
				} else {
					var sourceIndex = _this.uniqueSourceIndexByFilename[source.filename];
					mappings = source.content.getMappings(options.hires, sourceIndex, offsets, names);
				}

				return prefix + mappings;
			}).join('');

			return new SourceMap({
				file: options.file ? options.file.split(/[\/\\]/).pop() : null,
				sources: this.uniqueSources.map(function (source) {
					return options.file ? getRelativePath(options.file, source.filename) : source.filename;
				}),
				sourcesContent: this.uniqueSources.map(function (source) {
					return options.includeContent ? source.content : null;
				}),
				names: names,
				mappings: encoded
			});
		},
		getIndentString: function getIndentString() {
			var indentStringCounts = {};

			this.sources.forEach(function (source) {
				var indentStr = source.content.indentStr;

				if (indentStr === null) return;

				if (!indentStringCounts[indentStr]) indentStringCounts[indentStr] = 0;
				indentStringCounts[indentStr] += 1;
			});

			return Object.keys(indentStringCounts).sort(function (a, b) {
				return indentStringCounts[a] - indentStringCounts[b];
			})[0] || '\t';
		},
		indent: function indent(indentStr) {
			var _this2 = this;

			if (!arguments.length) {
				indentStr = this.getIndentString();
			}

			if (indentStr === '') return this; // noop

			var trailingNewline = !this.intro || this.intro.slice(-1) === '\n';

			this.sources.forEach(function (source, i) {
				var separator = source.separator !== undefined ? source.separator : _this2.separator;
				var indentStart = trailingNewline || i > 0 && /\r?\n$/.test(separator);

				source.content.indent(indentStr, {
					exclude: source.indentExclusionRanges,
					indentStart: indentStart //: trailingNewline || /\r?\n$/.test( separator )  //true///\r?\n/.test( separator )
				});

				// TODO this is a very slow way to determine this
				trailingNewline = source.content.toString().slice(0, -1) === '\n';
			});

			if (this.intro) {
				this.intro = indentStr + this.intro.replace(/^[^\n]/gm, function (match, index) {
					return index > 0 ? indentStr + match : match;
				});
			}

			return this;
		},
		prepend: function prepend(str) {
			this.intro = str + this.intro;
			return this;
		},
		toString: function toString() {
			var _this3 = this;

			var body = this.sources.map(function (source, i) {
				var separator = source.separator !== undefined ? source.separator : _this3.separator;
				var str = (i > 0 ? separator : '') + source.content.toString();

				return str;
			}).join('');

			return this.intro + body;
		},
		trimLines: function trimLines() {
			return this.trim('[\\r\\n]');
		},
		trim: function trim(charType) {
			return this.trimStart(charType).trimEnd(charType);
		},
		trimStart: function trimStart(charType) {
			var rx = new RegExp('^' + (charType || '\\s') + '+');
			this.intro = this.intro.replace(rx, '');

			if (!this.intro) {
				var source = void 0;
				var i = 0;

				do {
					source = this.sources[i];

					if (!source) {
						break;
					}

					source.content.trimStart(charType);
					i += 1;
				} while (source.content.toString() === ''); // TODO faster way to determine non-empty source?
			}

			return this;
		},
		trimEnd: function trimEnd(charType) {
			var rx = new RegExp((charType || '\\s') + '+$');

			var source = void 0;
			var i = this.sources.length - 1;

			do {
				source = this.sources[i];

				if (!source) {
					this.intro = this.intro.replace(rx, '');
					break;
				}

				source.content.trimEnd(charType);
				i -= 1;
			} while (source.content.toString() === ''); // TODO faster way to determine non-empty source?

			return this;
		}
	};

	function getSemis(str) {
		return new Array(str.split('\n').length).join(';');
	}

	MagicString.Bundle = Bundle;

	return MagicString;

}));
//# sourceMappingURL=magic-string.umd.js.map